@@ -4,78 +4,51 @@
 # This module is part of GitPython and is released under
 # the BSD License: http://www.opensource.org/licenses/bsd-license.php
 
-from git.exc import (
-    InvalidGitRepositoryError,
-    NoSuchPathError,
-    GitCommandError
-)
+from builtins import str
+from collections import namedtuple
+import logging
+import os
+import re
+import warnings
+
 from git.cmd import (
     Git,
     handle_process_output
 )
-from git.refs import (
-    HEAD,
-    Head,
-    Reference,
-    TagReference,
-)
-from git.objects import (
-    Submodule,
-    RootModule,
-    Commit
-)
-from git.util import (
-    Actor,
-    finalize_process
-)
-from git.index import IndexFile
-from git.config import GitConfigParser
-from git.remote import (
-    Remote,
-    add_progress,
-    to_progress_instance
-)
-
-from git.db import GitCmdObjectDB
-
-from gitdb.util import (
-    join,
-    isfile,
-    hex_to_bin
-)
-
-from .fun import (
-    rev_parse,
-    is_git_dir,
-    find_git_dir,
-    touch,
-)
 from git.compat import (
     text_type,
     defenc,
     PY3,
     safe_decode,
     range,
+    is_win,
 )
+from git.config import GitConfigParser
+from git.db import GitCmdObjectDB
+from git.exc import InvalidGitRepositoryError, NoSuchPathError, GitCommandError
+from git.index import IndexFile
+from git.objects import Submodule, RootModule, Commit
+from git.refs import HEAD, Head, Reference, TagReference
+from git.remote import Remote, add_progress, to_progress_instance
+from git.util import Actor, finalize_process, decygpath, hex_to_bin, expand_path
+import os.path as osp
 
-import os
-import sys
-import re
-from collections import namedtuple
+from .fun import rev_parse, is_git_dir, find_submodule_git_dir, touch, find_worktree_git_dir
+import gc
+import gitdb
 
-DefaultDBType = GitCmdObjectDB
-if sys.version_info[:2] < (2, 5):     # python 2.4 compatiblity
-    DefaultDBType = GitCmdObjectDB
-# END handle python 2.4
+try:
+    import pathlib
+except ImportError:
+    pathlib = None
+
+
+log = logging.getLogger(__name__)
 
 BlameEntry = namedtuple('BlameEntry', ['commit', 'linenos', 'orig_path', 'orig_linenos'])
 
 
-__all__ = ('Repo', )
-
-
-def _expand_path(p):
-    return os.path.abspath(os.path.expandvars(os.path.expanduser(p)))
+__all__ = ('Repo',)
 
 
 class Repo(object):
@@ -93,7 +66,12 @@ class Repo(object):
 
     'git_dir' is the .git repository directory, which is always set."""
     DAEMON_EXPORT_FILE = 'git-daemon-export-ok'
-    __slots__ = ("working_dir", "_working_tree_dir", "git_dir", "_bare", "git", "odb")
+
+    _git = None  # Must exist, or  __del__  will fail in case we raise on `__init__()`
+    working_dir = None
+    _working_tree_dir = None
+    git_dir = None
+    _common_dir = None
 
     # precompiled regex
     re_whitespace = re.compile(r'\s+')
@@ -110,7 +88,7 @@ class Repo(object):
     # Subclasses may easily bring in their own custom types by placing a constructor or type here
     GitCommandWrapperType = Git
 
-    def __init__(self, path=None, odbt=DefaultDBType, search_parent_directories=False):
+    def __init__(self, path=None, odbt=GitCmdObjectDB, search_parent_directories=False, expand_vars=True):
         """Create a new Repo instance
 
         :param path:
@@ -120,7 +98,11 @@ class Repo(object):
                 repo = Repo("/Users/mtrier/Development/git-python.git")
                 repo = Repo("~/Development/git-python.git")
                 repo = Repo("$REPOSITORIES/Development/git-python.git")
+                repo = Repo("C:\\Users\\mtrier\\Development\\git-python\\.git")
 
+            - In *Cygwin*, path may be a `'cygdrive/...'` prefixed path.
+            - If it evaluates to false, :envvar:`GIT_DIR` is used, and if this also evals to false,
+              the current-directory is used.
         :param odbt:
             Object DataBase type - a type which is constructed by providing
             the directory containing the database objects, i.e. .git/objects. It will
@@ -133,37 +115,68 @@ class Repo(object):
         :raise InvalidGitRepositoryError:
         :raise NoSuchPathError:
         :return: git.Repo """
-        epath = _expand_path(path or os.getcwd())
-        self.git = None  # should be set for __del__ not to fail in case we raise
+
+        epath = path or os.getenv('GIT_DIR')
+        if not epath:
+            epath = os.getcwd()
+        if Git.is_cygwin():
+            epath = decygpath(epath)
+
+        epath = epath or path or os.getcwd()
+        if not isinstance(epath, str):
+            epath = str(epath)
+        if expand_vars and ("%" in epath or "$" in epath):
+            warnings.warn("The use of environment variables in paths is deprecated" +
+                          "\nfor security reasons and may be removed in the future!!")
+        epath = expand_path(epath, expand_vars)
         if not os.path.exists(epath):
             raise NoSuchPathError(epath)
 
-        self.working_dir = None
-        self._working_tree_dir = None
-        self.git_dir = None
+        ## Walk up the path to find the `.git` dir.
+        #
         curpath = epath
-
-        # walk up the path to find the .git dir
         while curpath:
-            # ABOUT os.path.NORMPATH
+            # ABOUT osp.NORMPATH
             # It's important to normalize the paths, as submodules will otherwise initialize their
             # repo instances with paths that depend on path-portions that will not exist after being
             # removed. It's just cleaner.
             if is_git_dir(curpath):
-                self.git_dir = os.path.normpath(curpath)
+                self.git_dir = curpath
+                # from man git-config : core.worktree
+                # Set the path to the root of the working tree. If GIT_COMMON_DIR environment
+                # variable is set, core.worktree is ignored and not used for determining the
+                # root of working tree. This can be overridden by the GIT_WORK_TREE environment
+                # variable. The value can be an absolute path or relative to the path to the .git
+                # directory, which is either specified by GIT_DIR, or automatically discovered.
+                # If GIT_DIR is specified but none of GIT_WORK_TREE and core.worktree is specified,
+                # the current working directory is regarded as the top level of your working tree.
                 self._working_tree_dir = os.path.dirname(self.git_dir)
+                if os.environ.get('GIT_COMMON_DIR') is None:
+                    gitconf = self.config_reader("repository")
+                    if gitconf.has_option('core', 'worktree'):
+                        self._working_tree_dir = gitconf.get('core', 'worktree')
+                if 'GIT_WORK_TREE' in os.environ:
+                    self._working_tree_dir = os.getenv('GIT_WORK_TREE')
                 break
 
-            gitpath = find_git_dir(join(curpath, '.git'))
-            if gitpath is not None:
-                self.git_dir = os.path.normpath(gitpath)
+            dotgit = osp.join(curpath, '.git')
+            sm_gitpath = find_submodule_git_dir(dotgit)
+            if sm_gitpath is not None:
+                self.git_dir = osp.normpath(sm_gitpath)
+
+            sm_gitpath = find_submodule_git_dir(dotgit)
+            if sm_gitpath is None:
+                sm_gitpath = find_worktree_git_dir(dotgit)
+
+            if sm_gitpath is not None:
+                self.git_dir = expand_path(sm_gitpath, expand_vars)
                 self._working_tree_dir = curpath
                 break
 
             if not search_parent_directories:
                 break
-            curpath, dummy = os.path.split(curpath)
-            if not dummy:
+            curpath, tail = osp.split(curpath)
+            if not tail:
                 break
         # END while curpath
 
@@ -177,24 +190,69 @@ class Repo(object):
             # lets not assume the option exists, although it should
             pass
 
+        try:
+            common_dir = open(osp.join(self.git_dir, 'commondir'), 'rt').readlines()[0].strip()
+            self._common_dir = osp.join(self.git_dir, common_dir)
+        except (OSError, IOError):
+            self._common_dir = None
+
         # adjust the wd in case we are actually bare - we didn't know that
         # in the first place
         if self._bare:
             self._working_tree_dir = None
         # END working dir handling
 
-        self.working_dir = self._working_tree_dir or self.git_dir
-        self.git = self.GitCommandWrapperType(self.working_dir)
+        self.working_dir = self._working_tree_dir or self.common_dir
 
         # special handling, in special times
-        args = [join(self.git_dir, 'objects')]
+        args = [osp.join(self.common_dir, 'objects')]
         if issubclass(odbt, GitCmdObjectDB):
             args.append(self.git)
         self.odb = odbt(*args)
 
+    def _get_git(self):
+        working_dir = self._working_tree_dir or self.common_dir
+        if self._git:
+            if self._git._working_dir != expand_path(working_dir):
+                self.close()
+                self._git = None
+
+        if not self._git:
+            self._git = self.GitCommandWrapperType(working_dir)
+        return self._git
+
+    def _del_git(self):
+        if self._git:
+            self._git.clear_cache()
+            self._git = None
+            # Tempfiles objects on Windows are holding references to
+            # open files until they are collected by the garbage
+            # collector, thus preventing deletion.
+            # TODO: Find these references and ensure they are closed
+            # and deleted synchronously rather than forcing a gc
+            # collection.
+            if is_win:
+                gc.collect()
+            gitdb.util.mman.collect()
+            if is_win:
+                gc.collect()
+
+    git = property(fget=_get_git, fdel=_del_git)
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.close()
+
     def __del__(self):
-        if self.git:
-            self.git.clear_cache()
+        try:
+            self.close()
+        except Exception:
+            pass
+
+    def close(self):
+        del self.git
 
     def __eq__(self, rhs):
         if isinstance(rhs, Repo):
@@ -209,12 +267,14 @@ class Repo(object):
 
     # Description property
     def _get_description(self):
-        filename = join(self.git_dir, 'description')
-        return open(filename, 'rb').read().rstrip().decode(defenc)
+        filename = osp.join(self.git_dir, 'description')
+        with open(filename, 'rb') as fp:
+            return fp.read().rstrip().decode(defenc)
 
     def _set_description(self, descr):
-        filename = join(self.git_dir, 'description')
-        open(filename, 'wb').write((descr + '\n').encode(defenc))
+        filename = osp.join(self.git_dir, 'description')
+        with open(filename, 'wb') as fp:
+            fp.write((descr + '\n').encode(defenc))
 
     description = property(_get_description, _set_description,
                            doc="the project's description")
@@ -227,6 +287,13 @@ class Repo(object):
         """
         return self._working_tree_dir
 
+    @property
+    def common_dir(self):
+        """:return: The git dir that holds everything except possibly HEAD,
+        FETCH_HEAD, ORIG_HEAD, COMMIT_EDITMSG, index, and logs/ .
+        """
+        return self._common_dir or self.git_dir
+
     @property
     def bare(self):
         """:return: True if the repository is bare"""
@@ -370,18 +437,26 @@ class Repo(object):
     def _get_config_path(self, config_level):
         # we do not support an absolute path of the gitconfig on windows ,
         # use the global config instead
-        if sys.platform == "win32" and config_level == "system":
+        if is_win and config_level == "system":
             config_level = "global"
 
         if config_level == "system":
             return "/etc/gitconfig"
         elif config_level == "user":
-            config_home = os.environ.get("XDG_CONFIG_HOME") or os.path.join(os.environ.get("HOME", '~'), ".config")
-            return os.path.normpath(os.path.expanduser(join(config_home, "git", "config")))
+            config_home = os.environ.get("XDG_CONFIG_HOME") or osp.join(os.environ.get("HOME", '~'), ".config")
+            return osp.normpath(osp.expanduser(osp.join(config_home, "git", "config")))
         elif config_level == "global":
-            return os.path.normpath(os.path.expanduser("~/.gitconfig"))
+            return osp.normpath(osp.expanduser("~/.gitconfig"))
         elif config_level == "repository":
-            return os.path.normpath(join(self.git_dir, "config"))
+            try:
+                config_path = self.git.rev_parse("config", git_path=True)
+            except GitCommandError:
+                return osp.normpath(osp.join(self._common_dir or self.git_dir, "config"))
+            else:
+                if self.git._working_dir:
+                    return osp.normpath(osp.join(self.git._working_dir, config_path))
+                else:
+                    return config_path
 
         raise ValueError("Invalid configuration level: %r" % config_level)
 
@@ -396,8 +471,7 @@ class Repo(object):
         :param config_level:
             For possible values, see config_writer method
             If None, all applicable levels will be used. Specify a level in case
-            you know which exact file you whish to read to prevent reading multiple files for
-            instance
+            you know which file you wish to read to prevent reading multiple files.
         :note: On windows, system configuration cannot currently be read as the path is
             unknown, instead the global path will be used."""
         files = None
@@ -417,7 +491,7 @@ class Repo(object):
 
         :param config_level:
             One of the following values
-            system = sytem wide configuration file
+            system = system wide configuration file
             global = user level configuration file
             repository = configuration file for this repostory only"""
         return GitConfigParser(self._get_config_path(config_level), read_only=False)
@@ -457,22 +531,22 @@ class Repo(object):
     def iter_commits(self, rev=None, paths='', **kwargs):
         """A list of Commit objects representing the history of a given ref/commit
 
-        :parm rev:
+        :param rev:
             revision specifier, see git-rev-parse for viable options.
             If None, the active branch will be used.
 
-        :parm paths:
+        :param paths:
             is an optional path or a list of paths to limit the returned commits to
             Commits that do not contain that path or the paths will not be returned.
 
-        :parm kwargs:
+        :param kwargs:
             Arguments to be passed to git-rev-list - common ones are
             max_count and skip
 
         :note: to receive only commits between two named revisions, use the
             "revA...revB" revision specifier
 
-        :return ``git.Commit[]``"""
+        :return: ``git.Commit[]``"""
         if rev is None:
             rev = self.head.commit
 
@@ -491,7 +565,7 @@ class Repo(object):
             raise ValueError("Please specify at least two revs, got only %i" % len(rev))
         # end handle input
 
-        res = list()
+        res = []
         try:
             lines = self.git.merge_base(*rev, **kwargs).splitlines()
         except GitCommandError as err:
@@ -510,11 +584,11 @@ class Repo(object):
         return res
 
     def is_ancestor(self, ancestor_rev, rev):
-        """Check if a commit  is an ancestor of another
+        """Check if a commit is an ancestor of another
 
         :param ancestor_rev: Rev which should be an ancestor
         :param rev: Rev to test against ancestor_rev
-        :return: ``True``, ancestor_rev is an accestor to rev.
+        :return: ``True``, ancestor_rev is an ancestor to rev.
         """
         try:
             self.git.merge_base(ancestor_rev, rev, is_ancestor=True)
@@ -525,12 +599,12 @@ class Repo(object):
         return True
 
     def _get_daemon_export(self):
-        filename = join(self.git_dir, self.DAEMON_EXPORT_FILE)
-        return os.path.exists(filename)
+        filename = osp.join(self.git_dir, self.DAEMON_EXPORT_FILE)
+        return osp.exists(filename)
 
     def _set_daemon_export(self, value):
-        filename = join(self.git_dir, self.DAEMON_EXPORT_FILE)
-        fileexists = os.path.exists(filename)
+        filename = osp.join(self.git_dir, self.DAEMON_EXPORT_FILE)
+        fileexists = osp.exists(filename)
         if value and not fileexists:
             touch(filename)
         elif not value and fileexists:
@@ -545,41 +619,33 @@ class Repo(object):
         """The list of alternates for this repo from which objects can be retrieved
 
         :return: list of strings being pathnames of alternates"""
-        alternates_path = join(self.git_dir, 'objects', 'info', 'alternates')
+        alternates_path = osp.join(self.git_dir, 'objects', 'info', 'alternates')
 
-        if os.path.exists(alternates_path):
-            try:
-                f = open(alternates_path, 'rb')
+        if osp.exists(alternates_path):
+            with open(alternates_path, 'rb') as f:
                 alts = f.read().decode(defenc)
-            finally:
-                f.close()
             return alts.strip().splitlines()
         else:
-            return list()
+            return []
 
     def _set_alternates(self, alts):
         """Sets the alternates
 
-        :parm alts:
+        :param alts:
             is the array of string paths representing the alternates at which
             git should look for objects, i.e. /home/user/repo/.git/objects
 
         :raise NoSuchPathError:
         :note:
-            The method does not check for the existance of the paths in alts
+            The method does not check for the existence of the paths in alts
             as the caller is responsible."""
-        alternates_path = join(self.git_dir, 'objects', 'info', 'alternates')
+        alternates_path = osp.join(self.common_dir, 'objects', 'info', 'alternates')
         if not alts:
-            if isfile(alternates_path):
+            if osp.isfile(alternates_path):
                 os.remove(alternates_path)
         else:
-            try:
-                f = open(alternates_path, 'wb')
+            with open(alternates_path, 'wb') as f:
                 f.write("\n".join(alts).encode(defenc))
-            finally:
-                f.close()
-            # END file handling
-        # END alts handling
 
     alternates = property(_get_alternates, _set_alternates,
                           doc="Retrieve a list of alternates paths or set a list paths to be used as alternates")
@@ -604,7 +670,7 @@ class Repo(object):
             default_args.append(path)
         if index:
             # diff index against HEAD
-            if isfile(self.index.path) and \
+            if osp.isfile(self.index.path) and \
                     len(self.git.diff('--cached', *default_args)):
                 return True
         # END index handling
@@ -636,7 +702,7 @@ class Repo(object):
         return self._get_untracked_files()
 
     def _get_untracked_files(self, *args, **kwargs):
-        # make sure we get all files, no only untracked directores
+        # make sure we get all files, not only untracked directories
         proc = self.git.status(*args,
                                porcelain=True,
                                untracked_files=True,
@@ -644,7 +710,7 @@ class Repo(object):
                                **kwargs)
         # Untracked files preffix in porcelain mode
         prefix = "?? "
-        untracked_files = list()
+        untracked_files = []
         for line in proc.stdout:
             line = line.decode(defenc)
             if not line.startswith(prefix):
@@ -675,7 +741,7 @@ class Repo(object):
         Unlike .blame(), this does not return the actual file's contents, only
         a stream of BlameEntry tuples.
 
-        :parm rev: revision specifier, see git-rev-parse for viable options.
+        :param rev: revision specifier, see git-rev-parse for viable options.
         :return: lazy iterator of BlameEntry tuples, where the commit
                  indicates the commit to blame for the line, and range
                  indicates a span of line numbers in the resulting file.
@@ -684,11 +750,14 @@ class Repo(object):
         should get a continuous range spanning all line numbers in the file.
         """
         data = self.git.blame(rev, '--', file, p=True, incremental=True, stdout_as_string=False, **kwargs)
-        commits = dict()
+        commits = {}
 
         stream = (line for line in data.split(b'\n') if line)
         while True:
-            line = next(stream)  # when exhausted, casues a StopIteration, terminating this function
+            try:
+                line = next(stream)  # when exhausted, causes a StopIteration, terminating this function
+            except StopIteration:
+                return
             hexsha, orig_lineno, lineno, num_lines = line.split()
             lineno = int(lineno)
             num_lines = int(num_lines)
@@ -696,9 +765,12 @@ class Repo(object):
             if hexsha not in commits:
                 # Now read the next few lines and build up a dict of properties
                 # for this commit
-                props = dict()
+                props = {}
                 while True:
-                    line = next(stream)
+                    try:
+                        line = next(stream)
+                    except StopIteration:
+                        return
                     if line == b'boundary':
                         # "boundary" indicates a root commit and occurs
                         # instead of the "previous" tag
@@ -720,11 +792,17 @@ class Repo(object):
                            committed_date=int(props[b'committer-time']))
                 commits[hexsha] = c
             else:
-                # Discard the next line (it's a filename end tag)
-                line = next(stream)
-                tag, value = line.split(b' ', 1)
-                assert tag == b'filename', 'Unexpected git blame output'
-                orig_filename = value
+                # Discard all lines until we find "filename" which is
+                # guaranteed to be the last line
+                while True:
+                    try:
+                        line = next(stream)  # will fail if we reach the EOF unexpectedly
+                    except StopIteration:
+                        return
+                    tag, value = line.split(b' ', 1)
+                    if tag == b'filename':
+                        orig_filename = value
+                        break
 
             yield BlameEntry(commits[hexsha],
                              range(lineno, lineno + num_lines),
@@ -734,7 +812,7 @@ class Repo(object):
     def blame(self, rev, file, incremental=False, **kwargs):
         """The blame information for the given file at the given revision.
 
-        :parm rev: revision specifier, see git-rev-parse for viable options.
+        :param rev: revision specifier, see git-rev-parse for viable options.
         :return:
             list: [git.Commit, list: [<line>]]
             A list of tuples associating a Commit object with a list of lines that
@@ -744,8 +822,8 @@ class Repo(object):
             return self.blame_incremental(rev, file, **kwargs)
 
         data = self.git.blame(rev, '--', file, p=True, stdout_as_string=False, **kwargs)
-        commits = dict()
-        blames = list()
+        commits = {}
+        blames = []
         info = None
 
         keepends = True
@@ -840,7 +918,7 @@ class Repo(object):
         return blames
 
     @classmethod
-    def init(cls, path=None, mkdir=True, odbt=DefaultDBType, **kwargs):
+    def init(cls, path=None, mkdir=True, odbt=GitCmdObjectDB, expand_vars=True, **kwargs):
         """Initialize a git repository at the given path if specified
 
         :param path:
@@ -848,7 +926,7 @@ class Repo(object):
             or None in which case the repository will be created in the current
             working directory
 
-        :parm mkdir:
+        :param mkdir:
             if specified will create the repository directory if it doesn't
             already exists. Creates the directory with a mode=0755.
             Only effective if a path is explicitly given
@@ -858,13 +936,18 @@ class Repo(object):
             the directory containing the database objects, i.e. .git/objects.
             It will be used to access all object data
 
-        :parm kwargs:
+        :param expand_vars:
+            if specified, environment variables will not be escaped. This
+            can lead to information disclosure, allowing attackers to
+            access the contents of environment variables
+
+        :param kwargs:
             keyword arguments serving as additional options to the git-init command
 
         :return: ``git.Repo`` (the newly created repo)"""
         if path:
-            path = _expand_path(path)
-        if mkdir and path and not os.path.exists(path):
+            path = expand_path(path, expand_vars)
+        if mkdir and path and not osp.exists(path):
             os.makedirs(path, 0o755)
 
         # git command automatically chdir into the directory
@@ -873,110 +956,106 @@ class Repo(object):
         return cls(path, odbt=odbt)
 
     @classmethod
-    def _clone(cls, git, url, path, odb_default_type, progress, **kwargs):
+    def _clone(cls, git, url, path, odb_default_type, progress, multi_options=None, **kwargs):
         if progress is not None:
             progress = to_progress_instance(progress)
 
-        # special handling for windows for path at which the clone should be
-        # created.
-        # tilde '~' will be expanded to the HOME no matter where the ~ occours. Hence
-        # we at least give a proper error instead of letting git fail
-        prev_cwd = None
-        prev_path = None
         odbt = kwargs.pop('odbt', odb_default_type)
-        if os.name == 'nt':
-            if '~' in path:
-                raise OSError("Git cannot handle the ~ character in path %r correctly" % path)
 
-            # on windows, git will think paths like c: are relative and prepend the
-            # current working dir ( before it fails ). We temporarily adjust the working
-            # dir to make this actually work
-            match = re.match("(\w:[/\\\])(.*)", path)
-            if match:
-                prev_cwd = os.getcwd()
-                prev_path = path
-                drive, rest_of_path = match.groups()
-                os.chdir(drive)
-                path = rest_of_path
-                kwargs['with_keep_cwd'] = True
-            # END cwd preparation
-        # END windows handling
+        # when pathlib.Path or other classbased path is passed
+        if not isinstance(path, str):
+            path = str(path)
 
-        try:
-            proc = git.clone(url, path, with_extended_output=True, as_process=True,
-                             v=True, **add_progress(kwargs, git, progress))
-            if progress:
-                handle_process_output(proc, None, progress.new_message_handler(), finalize_process)
-            else:
-                (stdout, stderr) = proc.communicate()
-                finalize_process(proc, stderr=stderr)
-            # end handle progress
-        finally:
-            if prev_cwd is not None:
-                os.chdir(prev_cwd)
-                path = prev_path
-            # END reset previous working dir
-        # END bad windows handling
+        ## A bug win cygwin's Git, when `--bare` or `--separate-git-dir`
+        #  it prepends the cwd or(?) the `url` into the `path, so::
+        #        git clone --bare  /cygwin/d/foo.git  C:\\Work
+        #  becomes::
+        #        git clone --bare  /cygwin/d/foo.git  /cygwin/d/C:\\Work
+        #
+        clone_path = (Git.polish_url(path)
+                      if Git.is_cygwin() and 'bare' in kwargs
+                      else path)
+        sep_dir = kwargs.get('separate_git_dir')
+        if sep_dir:
+            kwargs['separate_git_dir'] = Git.polish_url(sep_dir)
+        multi = None
+        if multi_options:
+            multi = ' '.join(multi_options).split(' ')
+        proc = git.clone(multi, Git.polish_url(url), clone_path, with_extended_output=True, as_process=True,
+                         v=True, universal_newlines=True, **add_progress(kwargs, git, progress))
+        if progress:
+            handle_process_output(proc, None, progress.new_message_handler(), finalize_process, decode_streams=False)
+        else:
+            (stdout, stderr) = proc.communicate()
+            log.debug("Cmd(%s)'s unused stdout: %s", getattr(proc, 'args', ''), stdout)
+            finalize_process(proc, stderr=stderr)
 
         # our git command could have a different working dir than our actual
         # environment, hence we prepend its working dir if required
-        if not os.path.isabs(path) and git.working_dir:
-            path = join(git._working_dir, path)
+        if not osp.isabs(path) and git.working_dir:
+            path = osp.join(git._working_dir, path)
+
+        repo = cls(path, odbt=odbt)
+
+        # retain env values that were passed to _clone()
+        repo.git.update_environment(**git.environment())
 
         # adjust remotes - there may be operating systems which use backslashes,
         # These might be given as initial paths, but when handling the config file
         # that contains the remote from which we were clones, git stops liking it
         # as it will escape the backslashes. Hence we undo the escaping just to be
         # sure
-        repo = cls(os.path.abspath(path), odbt=odbt)
         if repo.remotes:
-            writer = repo.remotes[0].config_writer
-            writer.set_value('url', repo.remotes[0].url.replace("\\\\", "\\").replace("\\", "/"))
-            # PY3: be sure cleanup is performed and lock is released
-            writer.release()
+            with repo.remotes[0].config_writer as writer:
+                writer.set_value('url', Git.polish_url(repo.remotes[0].url))
         # END handle remote repo
         return repo
 
-    def clone(self, path, progress=None, **kwargs):
+    def clone(self, path, progress=None, multi_options=None, **kwargs):
         """Create a clone from this repository.
 
         :param path: is the full path of the new repo (traditionally ends with ./<name>.git).
         :param progress: See 'git.remote.Remote.push'.
+        :param multi_options: A list of Clone options that can be provided multiple times.  One
+            option per list item which is passed exactly as specified to clone.
+            For example ['--config core.filemode=false', '--config core.ignorecase',
+                         '--recurse-submodule=repo1_path', '--recurse-submodule=repo2_path']
         :param kwargs:
             * odbt = ObjectDatabase Type, allowing to determine the object database
               implementation used by the returned Repo instance
             * All remaining keyword arguments are given to the git-clone command
 
         :return: ``git.Repo`` (the newly cloned repo)"""
-        return self._clone(self.git, self.git_dir, path, type(self.odb), progress, **kwargs)
+        return self._clone(self.git, self.common_dir, path, type(self.odb), progress, multi_options, **kwargs)
 
     @classmethod
-    def clone_from(cls, url, to_path, progress=None, env=None, **kwargs):
+    def clone_from(cls, url, to_path, progress=None, env=None, multi_options=None, **kwargs):
         """Create a clone from the given URL
 
         :param url: valid git url, see http://www.kernel.org/pub/software/scm/git/docs/git-clone.html#URLS
         :param to_path: Path to which the repository should be cloned to
         :param progress: See 'git.remote.Remote.push'.
         :param env: Optional dictionary containing the desired environment variables.
+        :param multi_options: See ``clone`` method
         :param kwargs: see the ``clone`` method
         :return: Repo instance pointing to the cloned directory"""
         git = Git(os.getcwd())
         if env is not None:
             git.update_environment(**env)
-        return cls._clone(git, url, to_path, GitCmdObjectDB, progress, **kwargs)
+        return cls._clone(git, url, to_path, GitCmdObjectDB, progress, multi_options, **kwargs)
 
     def archive(self, ostream, treeish=None, prefix=None, **kwargs):
         """Archive the tree at the given revision.
 
-        :parm ostream: file compatible stream object to which the archive will be written as bytes
-        :parm treeish: is the treeish name/id, defaults to active branch
-        :parm prefix: is the optional prefix to prepend to each filename in the archive
-        :parm kwargs: Additional arguments passed to git-archive
+        :param ostream: file compatible stream object to which the archive will be written as bytes
+        :param treeish: is the treeish name/id, defaults to active branch
+        :param prefix: is the optional prefix to prepend to each filename in the archive
+        :param kwargs: Additional arguments passed to git-archive
 
             * Use the 'format' argument to define the kind of format. Use
               specialized ostreams to write any format supported by python.
             * You may specify the special **path** keyword, which may either be a repository-relative
-              path to a directory or file to place into the archive, or a list or tuple of multipe paths.
+              path to a directory or file to place into the archive, or a list or tuple of multiple paths.
 
         :raise GitCommandError: in case something went wrong
         :return: self"""
@@ -985,7 +1064,7 @@ class Repo(object):
         if prefix and 'prefix' not in kwargs:
             kwargs['prefix'] = prefix
         kwargs['output_stream'] = ostream
-        path = kwargs.pop('path', list())
+        path = kwargs.pop('path', [])
         if not isinstance(path, (tuple, list)):
             path = [path]
         # end assure paths is list
@@ -996,12 +1075,12 @@ class Repo(object):
     def has_separate_working_tree(self):
         """
         :return: True if our git_dir is not at the root of our working_tree_dir, but a .git file with a
-            platform agnositic symbolic link. Our git_dir will be whereever the .git file points to
+            platform agnositic symbolic link. Our git_dir will be wherever the .git file points to
         :note: bare repositories will always return False here
         """
         if self.bare:
             return False
-        return os.path.isfile(os.path.join(self.working_tree_dir, '.git'))
+        return osp.isfile(osp.join(self.working_tree_dir, '.git'))
 
     rev_parse = rev_parse
 
